from __future__ import annotations

import argparse
import contextlib
import io
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any

from argcomplete.completers import DirectoriesCompleter, FilesCompleter
from packaging.requirements import InvalidRequirement, Requirement

from pdm.compat import tomllib

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping

    from pdm.core import Core


_FILE_DESTS = {
    "ca_certs",
    "config",
    "dest",
    "filename",
    "lockfile",
    "output",
    "override",
    "path",
    "project_path",
    "python",
}
_DIRECTORY_DESTS = {"dest", "project_path"}
_GROUP_DESTS = {"exclude", "excluded_groups", "group", "groups", "include"}
_PACKAGE_PATHS = {
    ("list",),
    ("remove",),
    ("show",),
    ("update",),
}
_SHELLS = ("bash", "zsh", "fish", "powershell", "pwsh")


def _project_root(parsed_args: argparse.Namespace) -> Path:
    if project_path := getattr(parsed_args, "project_path", None):
        return Path(project_path).expanduser().resolve()
    current = Path.cwd()
    for directory in (current, *current.parents):
        if directory.joinpath("pyproject.toml").is_file():
            return directory
    return current


def _load_pyproject(parsed_args: argparse.Namespace) -> dict[str, object]:
    pyproject = _project_root(parsed_args) / "pyproject.toml"
    if not pyproject.is_file():
        return {}
    try:
        with pyproject.open("rb") as file:
            return tomllib.load(file)
    except (OSError, ValueError):
        return {}


def _groups(prefix: str, parsed_args: argparse.Namespace, **_: Any) -> Mapping[str, str]:
    data = _load_pyproject(parsed_args)
    project = data.get("project", {})
    tool = data.get("tool", {})
    pdm = tool.get("pdm", {}) if isinstance(tool, dict) else {}
    names = {"default"}
    if isinstance(project, dict):
        optional = project.get("optional-dependencies", {})
        if isinstance(optional, dict):
            names.update(map(str, optional))
    dependency_groups = data.get("dependency-groups", {})
    if isinstance(dependency_groups, dict):
        names.update(map(str, dependency_groups))
    if isinstance(pdm, dict):
        dev = pdm.get("dev-dependencies", {})
        if isinstance(dev, dict):
            names.update(map(str, dev))

    completed, separator, _partial = prefix.rpartition(",")
    value_prefix = f"{completed}{separator}" if separator else ""
    return {f"{value_prefix}{name}": "Dependency group" for name in sorted(names)}


def _requirement_name(requirement: object) -> str | None:
    if not isinstance(requirement, str):
        return None
    try:
        return Requirement(requirement).name
    except InvalidRequirement:
        match = re.match(r"[A-Za-z0-9][A-Za-z0-9._-]*", requirement)
        return match.group() if match else None


def _project_packages(parsed_args: argparse.Namespace) -> Mapping[str, str]:
    data = _load_pyproject(parsed_args)
    requirements: list[object] = []
    project = data.get("project", {})
    tool = data.get("tool", {})
    pdm = tool.get("pdm", {}) if isinstance(tool, dict) else {}
    if isinstance(project, dict):
        requirements.extend(project.get("dependencies", []) or [])
        optional = project.get("optional-dependencies", {})
        if isinstance(optional, dict):
            for values in optional.values():
                if isinstance(values, list):
                    requirements.extend(values)
    dependency_groups = data.get("dependency-groups", {})
    if isinstance(dependency_groups, dict):
        for values in dependency_groups.values():
            if isinstance(values, list):
                requirements.extend(values)
    if isinstance(pdm, dict):
        dev = pdm.get("dev-dependencies", {})
        if isinstance(dev, dict):
            for values in dev.values():
                if isinstance(values, list):
                    requirements.extend(values)
    names = {name for item in requirements if (name := _requirement_name(item))}
    return dict.fromkeys(sorted(names, key=str.lower), "Project dependency")


def _packages(parsed_args: argparse.Namespace, **_: Any) -> Mapping[str, str]:
    return _project_packages(parsed_args)


def _scripts(parsed_args: argparse.Namespace, **_: Any) -> Mapping[str, str]:
    data = _load_pyproject(parsed_args)
    tool = data.get("tool", {})
    pdm = tool.get("pdm", {}) if isinstance(tool, dict) else {}
    scripts = pdm.get("scripts", {}) if isinstance(pdm, dict) else {}
    if not isinstance(scripts, dict):
        return {}
    return {str(name): "Project script" for name in sorted(scripts)}


def _config_keys(**_: Any) -> Mapping[str, str]:
    from pdm.project.config import Config

    return {name: item.description for name, item in sorted(Config._config_map.items()) if item.should_show()}


def _venvs(core: Core, parsed_args: argparse.Namespace, **_: Any) -> Mapping[str, str]:
    from pdm.cli.commands.venv.utils import iter_venvs

    try:
        with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
            project = core.create_project(_project_root(parsed_args))
            return {name: str(venv.root) for name, venv in iter_venvs(project)}
    except Exception:
        return {}


def _shells(**_: Any) -> Iterable[str]:
    return _SHELLS


def _set_completer(action: argparse.Action, completer: Callable[..., object]) -> None:
    action.completer = completer  # type: ignore[attr-defined]


def configure_parser(core: Core) -> None:
    """Attach value completers to PDM's fully constructed argument parser."""
    files_completer = FilesCompleter()
    directories_completer = DirectoriesCompleter()

    def visit(parser: argparse.ArgumentParser, path: tuple[str, ...] = ()) -> None:
        for action in parser._actions:
            if isinstance(action, argparse._SubParsersAction):
                if not path:
                    _set_completer(action, _scripts)
                for name, subparser in action.choices.items():
                    visit(subparser, (*path, name))
            elif action.dest in _GROUP_DESTS:
                _set_completer(action, _groups)
            elif action.dest in {"use_venv", "env", "venv"}:
                _set_completer(action, lambda **kwargs: _venvs(core, **kwargs))
            elif action.dest == "key" and path == ("config",):
                _set_completer(action, _config_keys)
            elif action.dest == "script" and path == ("run",):
                _set_completer(action, _scripts)
            elif action.dest in {"package", "packages", "patterns"} and path in _PACKAGE_PATHS:
                _set_completer(action, _packages)
            elif action.dest == "shell" and path == ("completion",):
                _set_completer(action, _shells)
            elif action.dest in _FILE_DESTS:
                completer = directories_completer if action.dest in _DIRECTORY_DESTS else files_completer
                _set_completer(action, completer)

    visit(core.parser)


__all__ = ["configure_parser"]



from __future__ import annotations

import base64
import hashlib
from pathlib import Path
from typing import TYPE_CHECKING

from findpython import BaseProvider, PythonVersion

from pdm.exceptions import PdmUsageError
from pdm.models.venv import VirtualEnv

if TYPE_CHECKING:
    import sys
    from collections.abc import Iterable

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    from pdm.project import Project


def hash_path(path: str) -> str:
    """Generate a hash for the given path."""
    return base64.urlsafe_b64encode(hashlib.new("md5", path.encode(), usedforsecurity=False).digest()).decode()[:8]


def get_in_project_venv(root: Path) -> VirtualEnv | None:
    """Get the python interpreter path of venv-in-project"""
    for possible_dir in (".venv", "venv", "env"):
        venv = VirtualEnv.get(root / possible_dir)
        if venv is not None:
            return venv
    return None


def get_venv_prefix(project: Project) -> str:
    """Get the venv prefix for the project"""
    path = project.root
    name_hash = hash_path(path.as_posix())
    return f"{path.name}-{name_hash}-"


def get_venv_parent(project: Project) -> Path:
    """Get the parent directory holding the centrally-created venvs.

    ``venv.location`` may contain a `~` prefix, which must be expanded the same
    way here as it is when the venv is created, otherwise the venvs are created
    in one directory and looked up in another.
    """
    return Path(project.config["venv.location"]).expanduser()


def iter_venvs(project: Project) -> Iterable[tuple[str, VirtualEnv]]:
    """Return an iterable of venv paths associated with the project"""
    in_project_venv = get_in_project_venv(project.root)
    if in_project_venv is not None:
        yield "in-project", in_project_venv
    venv_prefix = get_venv_prefix(project)
    venv_parent = get_venv_parent(project)
    for path in venv_parent.glob(f"{venv_prefix}*"):
        ident = path.name[len(venv_prefix) :]
        venv = VirtualEnv.get(path)
        if venv is not None:
            yield ident, venv


def iter_central_venvs(project: Project) -> Iterable[tuple[str, Path]]:
    """Return an iterable of all managed venvs and their paths."""
    venv_parent = get_venv_parent(project)
    for venv in venv_parent.glob("*"):
        ident = venv.name
        yield ident, venv


class VenvProvider(BaseProvider):
    """A Python provider for project venv pythons"""

    def __init__(self, project: Project) -> None:
        self.project = project

    @classmethod
    def create(cls) -> Self | None:
        return None

    def find_pythons(self) -> Iterable[PythonVersion]:
        for _, venv in iter_venvs(self.project):
            yield PythonVersion(venv.interpreter, _interpreter=venv.interpreter, keep_symlink=True)


def get_venv_with_name(project: Project, name: str) -> VirtualEnv:
    all_venvs = dict(iter_venvs(project))
    try:
        return all_venvs[name]
    except KeyError:
        raise PdmUsageError(
            f"No virtualenv with key '{name}' is found, must be one of {list(all_venvs)}.\n"
            "You can create one with 'pdm venv create'.",
        ) from None



from __future__ import annotations

import dataclasses
import hashlib
import json
import re
from collections.abc import Mapping, Sequence
from datetime import date, datetime, time
from typing import TYPE_CHECKING, Any

from dep_logic.markers import AnyMarker, BaseMarker, MarkerUnion

from pdm.exceptions import PdmException, RequirementError
from pdm.models.requirements import (
    ALLOW_ANY,
    FileRequirement,
    NamedRequirement,
    Requirement,
    VcsRequirement,
    parse_requirement,
)
from pdm.utils import cd, get_file_hash, get_requirement_from_override, normalize_name, url_without_fragments

if TYPE_CHECKING:
    from pathlib import Path

    from pdm.models.repositories.lock import LockedRepository, Package
    from pdm.project import Project


_URL_RE = re.compile(r"[A-Za-z][A-Za-z0-9+.-]*://")
_OPAQUE_REQUIREMENT_PREFIX = "url-sha256:"
_DYNAMIC_LOCK_FIELDS = frozenset({"dependencies", "optional-dependencies", "requires-python", "version"})


def _to_builtin(value: Any, *, allow_temporal: bool = False) -> Any:
    if hasattr(value, "unwrap"):
        value = value.unwrap()
    if isinstance(value, Mapping):
        return {key: _to_builtin(item, allow_temporal=allow_temporal) for key, item in value.items()}
    if isinstance(value, (list, tuple)):
        return [_to_builtin(item, allow_temporal=allow_temporal) for item in value]
    if allow_temporal and isinstance(value, (datetime, date, time)):
        return value.isoformat()
    if value is None or isinstance(value, (bool, float, int, str)):
        return value
    raise TypeError(f"Unsupported lock input type: {type(value).__name__}")


def _fingerprint(value: str, prefix: str = "sha256:") -> str:
    return prefix + hashlib.sha256(value.encode("utf-8")).hexdigest()


def _requirement_input(requirement: Requirement) -> str:
    line = requirement.as_line()
    if isinstance(requirement, NamedRequirement) and not _URL_RE.search(line):
        return dataclasses.replace(requirement, specifier=ALLOW_ANY).as_line()
    if isinstance(requirement, FileRequirement) and not requirement.url and requirement.path is not None:
        line = json.dumps(
            {
                "editable": requirement.editable,
                "extras": sorted(requirement.extras or []),
                "marker": str(requirement.marker or ""),
                "name": requirement.project_name or "",
                "path": requirement.str_path,
                "subdirectory": requirement.subdirectory or "",
            },
            sort_keys=True,
        )
    return _fingerprint(line, _OPAQUE_REQUIREMENT_PREFIX)


def _sanitize_strings(value: Any) -> Any:
    if isinstance(value, Mapping):
        return {key: _sanitize_strings(item) for key, item in value.items()}
    if isinstance(value, list):
        return [_sanitize_strings(item) for item in value]
    if isinstance(value, str) and _URL_RE.search(value):
        return _fingerprint(value)
    return value


def _normalize_sources(sources: Sequence[Mapping[str, Any]]) -> list[dict[str, Any]]:
    result: list[dict[str, Any]] = []
    for source in sources:
        item = dict(source)
        item.pop("username", None)
        item.pop("password", None)
        if url := item.get("url"):
            item["url"] = _fingerprint(str(url))
        for key in ("include_packages", "exclude_packages"):
            if key in item:
                item[key] = sorted(item[key])
        result.append(_sanitize_strings(item))
    return result


def _project_lock_inputs(
    project: Project,
    include_package: bool = False,
    managed_paths: frozenset[Path] = frozenset(),
) -> dict[str, Any]:
    dependencies = project._resolve_dependencies()
    resolution = _to_builtin(project.pyproject.resolution, allow_temporal=True)
    resolution.pop("lock_inputs", None)
    with cd(project.root):
        groups = {
            group: sorted(_requirement_input(requirement) for requirement in requirements)
            for group, requirements in sorted(dependencies.items())
        }

    result: dict[str, Any] = {
        "groups": groups,
        "requires-python": str(project.python_requires),
        "resolution": _sanitize_strings(resolution),
        "sources": _normalize_sources(_to_builtin(project.pyproject.settings.get("source", []), allow_temporal=True)),
    }
    if include_package:
        metadata = project.pyproject.metadata
        dynamic = sorted(metadata.get("dynamic", []))
        result["package"] = {
            "build-system": _sanitize_strings(_to_builtin(project.pyproject.build_system, allow_temporal=True)),
            "distribution": project.is_distribution,
            "dynamic": dynamic,
            "name": metadata.get("name", ""),
            "version": metadata.get("version", ""),
        }
        if _DYNAMIC_LOCK_FIELDS.intersection(dynamic):
            result["volatile"] = True

    local_inputs: dict[str, Any] = {}
    for requirements in dependencies.values():
        for requirement in requirements:
            if not isinstance(requirement, FileRequirement) or requirement.absolute_path is None:
                continue
            path = requirement.absolute_path.resolve()
            if path in managed_paths:
                continue
            key = _requirement_input(requirement)
            if path.is_file():
                local_inputs[key] = {"file-sha256": get_file_hash(path)}
            elif path.is_dir():
                local_project = project.core.create_project(path)
                if local_project.pyproject.is_valid:
                    local_inputs[key] = _project_lock_inputs(
                        local_project,
                        include_package=True,
                        managed_paths=managed_paths | {path},
                    )
                else:
                    local_inputs[key] = {"volatile": True}
    if local_inputs:
        result["local"] = {key: local_inputs[key] for key in sorted(local_inputs)}
    return result


def build_lock_inputs(project: Project) -> dict[str, Any]:
    """Build the canonical project inputs that determine a lock resolution."""
    workspace_project = project.workspace_project or project
    root = workspace_project.root.resolve()
    members = (
        sorted(workspace_project.iter_members(), key=lambda path: path.relative_to(root).as_posix())
        if workspace_project.is_workspace_root
        else []
    )
    managed_paths = frozenset({root, *(member.resolve() for member in members)})
    result: dict[str, Any] = {
        "project": _project_lock_inputs(workspace_project, managed_paths=managed_paths),
    }
    if members:
        result["workspace"] = {
            member.relative_to(root).as_posix(): _project_lock_inputs(
                workspace_project.core.create_project(member),
                include_package=True,
                managed_paths=managed_paths,
            )
            for member in members
        }
    return result


def _contains_volatile(value: Any) -> bool:
    if isinstance(value, Mapping):
        return value.get("volatile") is True or any(_contains_volatile(item) for item in value.values())
    if isinstance(value, list):
        return any(_contains_volatile(item) for item in value)
    return False


def _requirements_match(
    current_lines: Sequence[str],
    locked_lines: Sequence[str],
) -> bool:
    if len(current_lines) != len(locked_lines):
        return False
    if not all(isinstance(line, str) for line in (*current_lines, *locked_lines)):
        return False
    return sorted(current_lines) == sorted(locked_lines)


def _project_inputs_match(
    current: Mapping[str, Any],
    locked: Mapping[str, Any],
) -> bool:
    current_metadata = {key: value for key, value in current.items() if key != "groups"}
    locked_metadata = {key: value for key, value in locked.items() if key != "groups"}
    if current_metadata != locked_metadata:
        return False

    current_groups = current.get("groups")
    locked_groups = locked.get("groups")
    if not isinstance(current_groups, Mapping) or not isinstance(locked_groups, Mapping):
        return False
    if current_groups.keys() != locked_groups.keys():
        return False
    return all(
        isinstance(current_groups[group], Sequence)
        and not isinstance(current_groups[group], str)
        and isinstance(locked_groups[group], Sequence)
        and not isinstance(locked_groups[group], str)
        and _requirements_match(current_groups[group], locked_groups[group])
        for group in current_groups
    )


def _project_overrides(project: Project) -> dict[str, Requirement] | None:
    result: dict[str, Requirement] = {}
    try:
        with cd(project.root):
            for name, value in project.pyproject.resolution.get("overrides", {}).items():
                requirement = parse_requirement(get_requirement_from_override(normalize_name(name), value))
                result[requirement.identify()] = requirement
    except (RequirementError, TypeError, ValueError):
        return None
    return result


def _effective_requirement(requirement: Requirement, overrides: Mapping[str, Requirement]) -> Requirement:
    effective = overrides.get(requirement.identify())
    if effective is None and requirement.key is not None:
        effective = overrides.get(requirement.key)
    return requirement if effective is None else effective


def _requirement_is_excluded(requirement: Requirement, excludes: frozenset[str]) -> bool:
    return requirement.identify() in excludes or (requirement.key is not None and requirement.key in excludes)


def _requirement_marker(requirement: Requirement) -> BaseMarker:
    return requirement.marker.inner if requirement.marker is not None else AnyMarker()


def _package_marker(package: Package) -> BaseMarker:
    marker = package.candidate.req.marker
    return marker.inner if marker is not None else AnyMarker()


def _applicable_regions(requirement: Requirement, repository: LockedRepository) -> list[BaseMarker] | None:
    if not repository.targets:
        return None
    requirement_marker = _requirement_marker(requirement)
    regions: list[BaseMarker] = []
    for target in repository.targets:
        region = requirement_marker & target.markers_with_python().inner
        if not region.is_empty():
            regions.append(region)
    return regions


def _package_matches_requirement_context(
    package: Package,
    requirement: Requirement,
    project: Project,
    group: str,
) -> bool:
    candidate = package.candidate
    if requirement.key is not None and normalize_name(candidate.name or "") != requirement.key:
        return False
    if candidate.req.groups and group not in candidate.req.groups:
        return False
    extras, dependency_groups = project.split_extras_groups([group])
    return package.marker.evaluate({"extras": set(extras), "dependency_groups": set(dependency_groups)})


def _context_packages(
    requirement: Requirement,
    project: Project,
    group: str,
    repository: LockedRepository,
    regions: Sequence[BaseMarker],
) -> list[Package]:
    return [
        package
        for package in repository.packages.values()
        if _package_matches_requirement_context(package, requirement, project, group)
        and any(not (region & _package_marker(package)).is_empty() for region in regions)
    ]


def _packages_cover_regions(packages: Sequence[Package], regions: Sequence[BaseMarker]) -> bool:
    if not packages:
        return False
    coverage = MarkerUnion.of(*(_package_marker(package) for package in packages))
    return all((region & coverage) == region for region in regions)


def _file_source_identity(requirement: FileRequirement, project: Project) -> tuple[Any, ...]:
    common = (requirement.editable, requirement.subdirectory or "")
    if requirement.absolute_path is not None:
        return ("path", requirement.absolute_path.resolve(), *common)
    if isinstance(requirement, VcsRequirement):
        repo = project.backend.expand_line(url_without_fragments(requirement.repo))
        return ("vcs", requirement.vcs, repo, requirement.ref or "", *common)
    url = project.backend.expand_line(url_without_fragments(requirement.get_full_url()))
    return ("url", url, *common)


def _requirement_matches(
    requirement: Requirement,
    effective: Requirement,
    project: Project,
    group: str,
    repository: LockedRepository,
) -> bool:
    regions = _applicable_regions(requirement, repository)
    if regions is None:
        return False
    if not regions:
        return True
    packages = _context_packages(requirement, project, group, repository, regions)
    if isinstance(effective, NamedRequirement):
        if not all(
            package.candidate.version is not None
            and effective.specifier.contains(package.candidate.version, prereleases=True)
            for package in packages
        ):
            return False
    elif isinstance(effective, FileRequirement):
        expected_source = _file_source_identity(effective, project)
        packages = [
            package
            for package in packages
            if isinstance(package.candidate.req, FileRequirement)
            and _file_source_identity(package.candidate.req, project) == expected_source
        ]
    else:  # pragma: no cover
        return False
    return _packages_cover_regions(packages, regions)


def _project_specifiers_match(
    project: Project,
    repository: LockedRepository,
    overrides: Mapping[str, Requirement],
    excludes: frozenset[str],
    managed_paths: frozenset[Path],
    groups: Sequence[str],
) -> bool:
    dependencies = project._resolve_dependencies(list(groups))
    for group, requirements in dependencies.items():
        for requirement in requirements:
            if _requirement_is_excluded(requirement, excludes):
                continue
            effective = _effective_requirement(requirement, overrides)
            if not _requirement_matches(requirement, effective, project, group, repository):
                return False
            if not isinstance(effective, FileRequirement) or effective.absolute_path is None:
                continue
            path = effective.absolute_path.resolve()
            if path in managed_paths or not path.is_dir():
                continue
            local_project = project.core.create_project(path)
            if not local_project.pyproject.is_valid:
                return False
            if not _project_specifiers_match(
                local_project,
                repository,
                overrides,
                excludes,
                managed_paths | {path},
                ["default"],
            ):
                return False
    return True


def _all_specifiers_match(project: Project, repository: LockedRepository) -> bool:
    root = project.root.resolve()
    members = (
        sorted(project.iter_members(), key=lambda path: path.relative_to(root).as_posix())
        if project.is_workspace_root
        else []
    )
    managed_paths = frozenset({root, *(member.resolve() for member in members)})
    overrides = _project_overrides(project)
    if overrides is None:
        return False
    excludes = frozenset(normalize_name(name) for name in project.pyproject.resolution.get("excludes", []))
    if not _project_specifiers_match(
        project,
        repository,
        overrides,
        excludes,
        managed_paths,
        project.lockfile.groups or ["default"],
    ):
        return False
    return all(
        _project_specifiers_match(
            project.core.create_project(member),
            repository,
            overrides,
            excludes,
            managed_paths,
            ["default"],
        )
        for member in members
    )


def lock_inputs_match(project: Project, locked_inputs: object) -> bool:
    """Return whether the current project is satisfied by the recorded lock inputs."""
    try:
        locked = _to_builtin(locked_inputs)
    except (TypeError, ValueError):
        return False
    if not isinstance(locked, Mapping):
        return False
    current = build_lock_inputs(project)
    if _contains_volatile(current) or _contains_volatile(locked):
        return False

    workspace_project = project.workspace_project or project
    if current != locked:
        current_project = current.get("project")
        locked_project = locked.get("project")
        if not isinstance(current_project, Mapping) or not isinstance(locked_project, Mapping):
            return False
        if not _project_inputs_match(current_project, locked_project):
            return False

        current_workspace = current.get("workspace", {})
        locked_workspace = locked.get("workspace", {})
        if not isinstance(current_workspace, Mapping) or not isinstance(locked_workspace, Mapping):
            return False
        if current_workspace.keys() != locked_workspace.keys():
            return False
        if not all(
            isinstance(current_workspace[path], Mapping)
            and isinstance(locked_workspace[path], Mapping)
            and _project_inputs_match(current_workspace[path], locked_workspace[path])
            for path in current_workspace
        ):
            return False

    try:
        return _all_specifiers_match(workspace_project, workspace_project.get_locked_repository())
    except (PdmException, TypeError, ValueError):
        return False
